/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UBoxTracker.cc
// Creation   : 07 07 1997
// Goal       : TBoxTracker class definition, see "UBoxTracker.h"
//
//  Nuria Oliver 1997, nuria@media.mit.edu

// MVC++ includes
#include "stdafx.h"
#include <math.h>

// unit include
#include "UBoxTracker.h"

// project includes
#include "UObjAlloc.h"
#include "UEVID30.h"

////////// static data members initialisation

static const float  kBTDefaultTheoRatio           = 0.7f;
// default theorical ratio width of the box over height of the box

static const float  kBTDefaultTheoRatioThreshold  = 0.3f;
// default threshold over wich the tracked box is no more a face

static const int    kBTValidityRange              = 2;
// number of pixel from the border to validate estimation

//
//    Init
//

ErrorNumber TBoxTracker::Init ()
{
	ErrorNumber     err = kErrNoErr;
	TEVID30*        cam;
	// initial kalman time deviation setting
	float           tmpKalmanTimeDev[2] = {20.0, 20.0};
	float			  tmpKalmanCov[4] = { .1f, 0.01f, 0.01f, 0.1f};
	
	// init data members
	m_fRatio              = 0.6f;
	m_fRatioThreshold     = kBTDefaultTheoRatioThreshold;
	m_fXThreshold         = 0.0;
	m_fYThreshold         = 0.0;
	m_fP                  = 1.0;
	m_fD                  = 1.0;
	m_bCamOn              = FALSE;
	m_bZoomOn				= FALSE;
	m_fTheoRatio          = kBTDefaultTheoRatio;
	m_fTheoRatioThreshold = kBTDefaultTheoRatioThreshold;
	
	m_pDefaultCam		    = (TCamControl*)NULL;
	m_pCamera				= (TCamControl*)NULL;
	m_pTimer				= (TTimer*)NULL;
	
	m_fBoundsXCenter  = 0.0;
	m_fBoundsYCenter  = 0.0;
	m_fCurrentXSpeed  = 0.0;
	m_fCurrentYSpeed  = 0.0;
	m_fCurrentXErr    = 0.0;
	m_fCurrentYErr    = 0.0;
	m_fCurrentZoom    = 0.0;
	m_bCurrentPanic   = FALSE;
	m_bounds.SetRectEmpty();	
	
	// Init the kalman
	m_speedKalman.Init();
	m_speedKalman.SetTimeDeviation( tmpKalmanTimeDev );
	m_speedKalman.SetObsCovariance( tmpKalmanCov );
	
	// get access to camera
	err = AllocAndInit (cam);
	
	if (err == kErrNoErr)
	{
		
		m_pDefaultCam = (TCamControl*)cam;
		m_pCamera     = m_pDefaultCam;
		
		// move camera to canonical position
		m_pCamera->Home ();
		m_pCamera->SetZoomSpeed (-2);
		
		// alloc and init timer
		err = AllocAndInit (m_pTimer);	  
	}
	
	// error handling
	if (err != kErrNoErr)
		this->Dispose ();
	
	return err;
}



//
//    Dispose
//

void TBoxTracker::Dispose ()
{
	// free the timer
	TestDisposeAndFree (m_pTimer);
	TestDisposeAndFree (m_pCamera);
	// dispose the kalman 
	m_speedKalman.Dispose();
}



//
//    SetCamera
//

void TBoxTracker::SetCamera (TCamControl* cam)
{
	m_pCamera   = cam;
	m_pCamera->Reset ();
}




//
//    SetCameraOnOff
//

void TBoxTracker::SetCameraOnOff (BOOL mode)
{
	if (mode)
		m_bCamOn = TRUE;
	else
		if (m_bCamOn)
		{
			m_bCamOn = FALSE;
			m_pCamera->SetPanSpeed (0);
			m_pCamera->SetTiltSpeed (0);
		}  
}




//
//    SetZoomOnOff
//

void TBoxTracker::SetZoomOnOff (BOOL mode)
{
	if (mode)
		m_bZoomOn = TRUE;
	else
		if (m_bZoomOn)
		{
			m_bZoomOn = FALSE;
			m_pCamera->SetZoomSpeed (0);
		}  
}




//
//    SetParameters
//

void TBoxTracker::SetParameters  (CRect  *bounds,
                                  float   heightRatio,
                                  float   p,
                                  float   d,
                                  float   xThreshold,
                                  float   yThreshold,
                                  float   ratioThreshold,
                                  float   theoRatio,
                                  float   theoRatioThreshold)
{
	if (bounds != (CRect*)NULL)
	{
		m_bounds = *bounds;
		CPoint       cp( m_bounds.CenterPoint() );
		m_fBoundsXCenter  = (float)(cp.x);
		m_fBoundsYCenter  = (float)(cp.y);
	}
	
	if (heightRatio >= 0)
		m_fRatio = heightRatio;
	
	if (p >= 0)
		m_fP = p;
	
	if (d >= 0)
		m_fD = d;
	
	if (xThreshold >= 0)
		m_fXThreshold = xThreshold;
	
	if (yThreshold >= 0)
		m_fYThreshold = yThreshold;
	
	if (ratioThreshold >= 0)
		m_fRatioThreshold = ratioThreshold;
	
	if (theoRatio >= 0)
		m_fTheoRatio = theoRatio;
	
	if (theoRatioThreshold >= 0)
		m_fTheoRatioThreshold = theoRatioThreshold;
}



//
//    GetParameters
//

void TBoxTracker::GetParameters  (CRect*  bounds,
                                  float*  heightRatio,
                                  float*  p,
                                  float*  d,
                                  float*  xThreshold,
                                  float*  yThreshold,
                                  float*  ratioThreshold,
                                  float*  theoRatio,
                                  float*  theoRatioThreshold)
{
	if (bounds != (CRect*)NULL)
		*bounds = m_bounds;
	
	if (heightRatio != (float*)NULL)
		*heightRatio = m_fRatio;
	
	if (p != (float*)NULL)
		*p = m_fP;
	
	if (d != (float*)NULL)
		*d = m_fD;
	
	if (xThreshold != (float*)NULL)
		*xThreshold = m_fXThreshold;
	
	if (yThreshold != (float*)NULL)
		*yThreshold = m_fYThreshold;
	
	if (ratioThreshold != (float*)NULL)
		*ratioThreshold = m_fRatioThreshold;
	
	if (theoRatio != (float*)NULL)
		*theoRatio = m_fTheoRatio;
	
	if (theoRatioThreshold != (float*)NULL)
		*theoRatioThreshold = m_fTheoRatioThreshold;
}




//
//    Update
//

BOOL TBoxTracker::Update (CRect box)
{
	float			estimX, estimY;
	float			error, errorD;
	float			kalmanParams[2];
	float			kalmanObsCov[4];
	float         kalmanTimeDev[2];
	
	// this array has the error and the errorD
	UINT			gap;
	int			zoomPos;
	float			zoomFactor;
	BOOL			inField;
	
	this->EstimatePosition (box, estimX, estimY, m_bCurrentPanic,
		m_fCurrentZoom, inField);
	
	if (m_bCamOn)
	{
		gap             = m_pTimer->Gap ();
		zoomPos         = (int) m_pCamera->GetZoomPosition ();
		zoomFactor      =   (float) 4.295184E-4f * zoomPos * zoomPos
			- 1.122835E-2f * zoomPos
			+ 1.18182f;
		
		error           = m_fBoundsXCenter - estimX;
		errorD          = (float)(error - m_fCurrentXErr) / gap;
		
		m_fCurrentXSpeed  = (m_fP*error + m_fD*errorD) / zoomFactor;
		kalmanParams[0]	  = m_fCurrentXSpeed;
		
		m_fCurrentXErr    = error;
		
		error           = m_fBoundsYCenter - estimY;
		errorD          = (float)(error - m_fCurrentYErr) / gap;
		
		m_fCurrentYSpeed  = (m_fP*error + m_fD*errorD) / zoomFactor;
		kalmanParams[1]	  = m_fCurrentYSpeed;
		
		m_fCurrentYErr    = error;
		
		// Before updating, compute the new covariance matrix
		// for the Y measurement. We have to fit a curve to
		// the observation covariance, proportional to the error in
		// the Y dimension
		m_speedKalman.GetObsCovariance( kalmanObsCov );
		m_speedKalman.GetTimeDeviation( kalmanTimeDev );
		
		//kalmanObsCov[3] = (kalmanTimeDev[1]*(kalmanObsCov[0]/kalmanTimeDev[0]))*(1+fabs(m_fCurrentYErr));
		m_speedKalman.SetObsCovariance( kalmanObsCov );
		
		// Update the speed kalman filter
		m_speedKalman.Update( kalmanParams, kalmanParams );
		// Set the smoothed values back to the data members
		m_fCurrentXSpeed = kalmanParams[0];
		m_fCurrentYSpeed = kalmanParams[1];
		
		if (fabs(m_fCurrentXSpeed) > m_fXThreshold)
			m_pCamera->SetPanSpeed   (m_fCurrentXSpeed);
		else
			m_pCamera->SetPanSpeed   (0.0);
		
		if (fabs(m_fCurrentYSpeed) > m_fYThreshold)
			m_pCamera->SetTiltSpeed  (m_fCurrentYSpeed);
		else
			m_pCamera->SetTiltSpeed  (0.0);
	}
	
	if (m_bZoomOn)
	{
		// manage zoom
		if (m_bCurrentPanic == TRUE)
			m_pCamera->SetZoomSpeed (-2);
		else
			if (m_fCurrentZoom - m_fRatio > m_fRatioThreshold)
				m_pCamera->SetZoomSpeed (-1);
			else
				if (m_fCurrentZoom - m_fRatio < -m_fRatioThreshold)
					m_pCamera->SetZoomSpeed (1);
				else
					m_pCamera->SetZoomSpeed (0);
	}
	
	return (BOOL )((!inField) ||
		(fabs ((float)box.Width () / (float)box.Height ()
		- m_fTheoRatio) < m_fTheoRatioThreshold));
}




//
//    EstimatePosition
//

void TBoxTracker::EstimatePosition (CRect&    box,
                                    float&    estimX,
                                    float&    estimY,
                                    BOOL&	  panic,
                                    float&    zoomState,
                                    BOOL&	  inField)
{
	//float   estMidWidth   = (float)(box.Width() * m_fTheoRatio);
	//float   estMidHeight  = (float)(abs(box.Height()) / m_fTheoRatio);
	float   estMidWidth   = (float)(box.Width() * .5);
	float   estMidHeight  = (float)(abs(box.Height()) * .5);
	CPoint  cp( box.CenterPoint() );
	float   centerX       = (float)(cp.x);
	float   centerY       = (float)(cp.y);
	
	
	panic   = TRUE;
	inField = FALSE;
	
	if (box.top - m_bounds.top > kBTValidityRange)
		// top valid
		if (m_bounds.bottom - box.bottom > kBTValidityRange)
			// bottom valid
		{
			// base the zoom on height
			zoomState = (float)(box.bottom - box.top) /
				(float)(m_bounds.bottom - m_bounds.top);
			
			if (box.left - m_bounds.left > kBTValidityRange)
				// left valid
				if (m_bounds.right - box.right > kBTValidityRange)
					// right valid
				{
					// all sides valid
					estimX	= centerX;
					estimY	= centerY;
					panic		= FALSE;
					inField	= TRUE;
				}
				else
					// right invalid
				{
					// all but right valid
					estimX	= (float)box.left + estMidWidth;
					estimY	= centerY;
					panic		= FALSE;
				}
				else
					// left invalid
					if (m_bounds.right - box.right > kBTValidityRange)
						// right valid
					{
						// all but left valid
						estimX	= (float)box.right - estMidWidth;
						estimY	= centerY;
						panic		= FALSE;
					}
					else
						// right invalid
					{
						// right and left invalid
						estimX  = m_fBoundsXCenter;
						estimY  = centerY;
					}
		}
		else
			// bottom invalid
		{
			// base the zoom on width
			zoomState = (float)(box.right - box.left) / m_fTheoRatio /
				(float)(m_bounds.bottom - m_bounds.top);
			
			if (box.left - m_bounds.left > kBTValidityRange)
				// left valid
				if (m_bounds.right - box.right > kBTValidityRange)
					// right valid
				{
					// all but bottom valid
					estimX = centerX;
					estimY = box.top + estMidHeight;
					panic = FALSE;
				}
				else
					// right invalid
				{
					// right and bottom invalid
					estimX = (float) m_bounds.right;
					estimY = (float) m_bounds.bottom;
				}
				else
					// left invalid
					if (m_bounds.right - box.right > kBTValidityRange)
						// right valid
					{
						// left and bottom invalid
						estimX = (float) m_bounds.left;
						estimY = (float) m_bounds.bottom;
					}
					else
						// right invalid
					{
						// only top valid
						estimX = (float) m_fBoundsXCenter;
						estimY = (float) m_bounds.bottom;
					}
		}
		else
			// top invalid
		{
			// base the zoom on width
			zoomState = (float)(box.right - box.left) / m_fTheoRatio /
                (float)(m_bounds.bottom - m_bounds.top);
			
			if (m_bounds.bottom - box.bottom > kBTValidityRange)
				// bottom valid
				if (box.left - m_bounds.left > kBTValidityRange)
					// left valid
					if (m_bounds.right - box.right > kBTValidityRange)
						// right valid
					{
						// all but top valid
						estimX = centerX;
						estimY = box.bottom - estMidHeight;
						panic = FALSE;
					}
					else
						// right invalid
					{
						// right top invalid
						estimX = (float) m_bounds.right;
						estimY = (float) m_bounds.top;
					}
					else
						// left invalid
						if (m_bounds.right - box.right > kBTValidityRange)
							// right valid
						{
							// top and left invalid
							estimX = (float) m_bounds.left;
							estimY = (float) m_bounds.top;
						}
						else
							// right invalid
						{
							// only bottom valid
							estimX  = (float) m_fBoundsXCenter;
							estimY  = (float) m_bounds.top;
						}
						else
							// bottom invalid
							if (box.left - m_bounds.left > kBTValidityRange)
								// left valid
								if (m_bounds.right - box.right > kBTValidityRange)
									// right valid
								{
									// top and bottom invalid
									estimX = centerX;
									estimY = m_fBoundsYCenter;
								}
								else
									// right invalid
								{
									// only left valid
									estimX = (float) m_bounds.right;
									estimY = (float) m_fBoundsYCenter;
								}
								else
									// left invalid
									if (m_bounds.right - box.right > kBTValidityRange)
										// right valid
									{
										// only right valid
										estimX = (float) m_bounds.left;
										estimY = (float) m_fBoundsYCenter;
									}
									else
										// right invalid
									{
										// everything out of field of view
										estimX = m_fBoundsXCenter;
										estimY = m_fBoundsYCenter;
									}
		}
}
