/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UBoxTracker.h
// Creation   : 07 07 1997
// Goal       : declares the TBoxTracker class, that drives a camera to track
//              a rigid rectabgle in a rectangle frame view
//
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UBOXTRACKER_H
#define _UBOXTRACKER_H

// standard includes

// project includes
#include "UErrors.h"
#include "UTimer.h"
#include "UCamControl.h"
#include "UKalman.h"

#ifndef M_PI
#define M_PI	3.14159265359
#endif

class TBoxTracker
{
public:
	
    ErrorNumber     Init ();
    // initialises the box tracker and uses an EVID30 camera driver
    // by default
	
    void            Dispose ();
	
    void            SetCamera (TCamControl* cam);
	
    void            SetCameraOnOff (BOOL mode);
    // switch on or off camera tracking
	
    inline void     SetCameraHome ();
    // move camera to home position
	
    void            SetZoomOnOff (BOOL mode);
    // switch on or off camera zooming
	
    inline void     SetZoomWide ();
    // Send camera to wider position
	
    inline BOOL		GetCameraOnOff ();
    // get current status of camera tracking
	
    inline BOOL		GetZoomOnOff ();
    // get current status of camera zooming
	
    void            SetParameters  (CRect*  bounds = (CRect*)NULL,
		float   heightRatio = -1,
		float   p = -1,
		float   d = -1,
		float   xThreshold = -1,
		float   yThreshold = -1,
		float   ratioThreshold = -1,
		float   theoRatio = -1,
		float   theoRatioThreshold = -1);
    // every parameter >= 0 (or != NULL for pointers) will be used to
    // set the internal parameters of the tracker.
    //    bounds      is the extent of the field of view
    //    heightRatio is the desired ratio: height of the box / height of the 
    //                live window
    //    p           is the proportional coefficient of the error on speed
    //    d           is the proportional coefficient of the time derivative
    //                error on speed
    //    xTreshold   is the threshold under which there will be no pan rotation
    //    yTreshold   is the threshold under which there will be no tilt rot.
    //    ratioThreshold  is the threshold under which there will be no zooming
    //    theoRatio   is the theorical ratio of aspect (width/height) of a face
    //    theoRatioThreshold  is the error ion ratio over which the box is no
    //                more considered as a face
	
    void            GetParameters  (CRect*  bounds = (CRect*)NULL,
		float*  heightRatio = (float*)NULL,
		float*  p = (float*)NULL,
		float*  d = (float*)NULL,
		float*  xThreshold = (float*)NULL,
		float*  yThreshold = (float*)NULL,
		float*  ratioThreshold =(float*)NULL,
		float*  theoRatio =(float*)NULL,
		float*  theoRatioThreshold =(float*)NULL);
    // any parameter that is not NULL will be set acording to the internal
    // state of the tracker.
	
    BOOL         Update (CRect box);
    // compute tracking commands from new estimation of tracked box position
    // return false only if the box entierly in the field of view AND it
    // doesn't have the correct shape.
	
    inline
		void            GetState (float& xErr, float& yErr,
		float& panSpeed, float& tiltSpeed,
		float& zoomRatio,
		BOOL& panic);
	
    inline
		void            GetCameraPosition (float& pan, float& tilt, float& zoom);
    inline
		void            GetCameraRadPosition (float& pan, 
		float& tilt, 
		float& zoom);
	
protected:
	
    CRect           m_bounds;
    float           m_fRatio;
    float           m_fRatioThreshold;
    float           m_fXThreshold;
    float           m_fYThreshold;
    float           m_fP;
    float           m_fD;
    float           m_fTheoRatio;
    float           m_fTheoRatioThreshold;
    BOOL			m_bCamOn;
    BOOL			m_bZoomOn;
	TKalman			m_speedKalman;  // kalman for the error and its derivative
	
    TCamControl*    m_pDefaultCam;
    TCamControl*    m_pCamera;
    TTimer*         m_pTimer;
	
    float           m_fBoundsXCenter;
    float           m_fBoundsYCenter;
    float           m_fCurrentXSpeed;
    float           m_fCurrentYSpeed;
    float           m_fCurrentXErr;
    float           m_fCurrentYErr;
    float           m_fCurrentZoom;
    BOOL			m_bCurrentPanic;
	
	
    void            EstimatePosition (CRect&    box,
		float&    estimX,
		float&    estimY,
		BOOL&		panic,
		float&    zoomState,
		BOOL&		inField);
};





////////// inline function definitions

//
//    GetCameraOnOff
//

inline BOOL TBoxTracker::GetCameraOnOff ()
{
	return m_bCamOn;
}




//
//    SetCameraHome
//

inline void TBoxTracker::SetCameraHome ()
{
	m_pCamera->Home ();
}




//
//    GetZoomOnOff
//

inline BOOL TBoxTracker::GetZoomOnOff ()
{
	return m_bZoomOn;
}



//
//    SetZoomWide
//

inline void TBoxTracker::SetZoomWide ()
{
	m_pCamera->SetZoomSpeed (-2);
}




//
//    GetState
//

inline
void TBoxTracker::GetState (float& xErr, float& yErr,
                            float& panSpeed, float& tiltSpeed,
                            float& zoomRatio,
                            BOOL& panic)
{
	xErr      = m_fCurrentXErr;
	yErr      = m_fCurrentYErr;
	panSpeed  = m_fCurrentXSpeed;
	tiltSpeed = m_fCurrentYSpeed;
	zoomRatio = m_fCurrentZoom;
	panic     = m_bCurrentPanic;
}




//
//    GetCameraPosition
//

void TBoxTracker::GetCameraPosition (float& pan, float& tilt, float& zoom)
{
	pan   = m_pCamera->GetPanPosition ();
	tilt  = m_pCamera->GetTiltPosition ();
	zoom  = m_pCamera->GetZoomPosition ();
}

//
//    GetCameraRadPosition
//
//    Same as above but in radians the angles

void TBoxTracker::GetCameraRadPosition (float& pan, float& tilt, float& zoom)
{
	pan   = (float) (m_pCamera->GetPanPosition ()*M_PI)/180;
	tilt  = (float) (m_pCamera->GetTiltPosition ()*M_PI)/180;
	zoom  = m_pCamera->GetZoomPosition ();
}


#endif

